@@ -1,11 +1,11 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /cvsroot/pgpool/pgpool-II/pool_timestamp.c,v 1.7 2010/02/17 14:00:14 t-ishii Exp $
+ * $Header: /cvsroot/pgpool/pgpool-II/pool_timestamp.c,v 1.14 2010/09/06 08:33:32 kitagawa Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL
  * written by Tatsuo Ishii
  *
- * Copyright (c) 2003-2009	PgPool Global Development Group
+ * Copyright (c) 2003-2010	PgPool Global Development Group
  *
  * Permission to use, copy, modify, and distribute this software and
  * its documentation for any purpose and without fee is hereby
@@ -25,6 +25,9 @@
 
 #include "pool.h"
 #include "pool_timestamp.h"
+#include "pool_relcache.h"
+#include "pool_select_walker.h"
+#include "pool_config.h"
 #include "parser/parsenodes.h"
 #include "parser/gramparse.h"
 #include "parser/pool_memory.h"
@@ -107,15 +110,44 @@ ts_unregister_func(void *data)
 static TSRel*
 relcache_lookup(TSRewriteContext *ctx)
 {
-#define ATTRDEFQUERY "SELECT attname, coalesce(d.adsrc = 'now()' OR d.adsrc LIKE '%%''now''::text%%', false)" \
+#define ATTRDEFQUERY "SELECT attname, coalesce((d.adsrc = 'now()' OR d.adsrc LIKE '%%''now''::text%%')" \
+	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
+	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
+	" a.atttypid = 'date'::regtype::oid OR" \
+	" a.atttypid = 'time'::regtype::oid OR" \
+	" a.atttypid = 'time with time zone'::regtype::oid)" \
+    " , false)" \
 	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
 	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
 	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.relname = '%s'" \
 	" ORDER BY a.attnum"
 
+#define ATTRDEFQUERY2 "SELECT attname, coalesce((d.adsrc = 'now()' OR d.adsrc LIKE '%%''now''::text%%')" \
+	" AND (a.atttypid = 'timestamp'::regtype::oid OR" \
+	" a.atttypid = 'timestamp with time zone'::regtype::oid OR" \
+	" a.atttypid = 'date'::regtype::oid OR" \
+	" a.atttypid = 'time'::regtype::oid OR" \
+	" a.atttypid = 'time with time zone'::regtype::oid)" \
+    " , false)" \
+	" FROM pg_catalog.pg_class c, pg_catalog.pg_attribute a " \
+	" LEFT JOIN pg_catalog.pg_attrdef d ON (a.attrelid = d.adrelid AND a.attnum = d.adnum)" \
+	" WHERE c.oid = a.attrelid AND a.attnum >= 1 AND a.attisdropped = 'f' AND c.oid = pgpool_regclass('%s')" \
+	" ORDER BY a.attnum"
+
+	char *query;
+
+	if (pool_has_pgpool_regclass())
+	{
+		query = ATTRDEFQUERY2;
+	}
+	else
+	{
+		query = ATTRDEFQUERY;
+	}
+
 	if (!ts_relcache)
 	{
-		ts_relcache = pool_create_relcache(MAX_RELCACHE, ATTRDEFQUERY, ts_register_func, ts_unregister_func, false);
+		ts_relcache = pool_create_relcache(MAX_RELCACHE, query, ts_register_func, ts_unregister_func, false);
 
 		if (ts_relcache == NULL)
 		{
@@ -528,7 +560,7 @@ rewrite_timestamp_update(UpdateStmt *u_stmt, TSRewriteContext *ctx)
  */
 char *
 rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
-	   	bool rewrite_to_params, Portal *portal)
+				  bool rewrite_to_params, PreparedStatement *pstmt)
 {
 	TSRewriteContext	ctx;
 	Node			*stmt;
@@ -598,11 +630,11 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 		rewrite = ctx.rewrite;
 
 		/* add params */
-		if (portal)
+		if (pstmt)
 		{
 			int		i;
 
-			for (i = 0; i < portal->num_tsparams; i++)
+			for (i = 0; i < pstmt->num_tsparams; i++)
 			{
 				e_stmt->params = lappend(e_stmt->params, ctx.ts_const);
 				rewrite = true;
@@ -615,7 +647,7 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 	if (!rewrite)
 		return NULL;
 
-	if (ctx.rewrite_to_params && portal)
+	if (ctx.rewrite_to_params && pstmt)
 	{
 		ListCell	*lc;
 		int			 num = ctx.num_params + 1;
@@ -628,7 +660,7 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 		}
 
 		/* save to portal */
-		portal->num_tsparams = list_length(ctx.params);
+		pstmt->num_tsparams = list_length(ctx.params);
 
 		/* add param type */
 		if (IsA(node, PrepareStmt))
@@ -636,7 +668,7 @@ rewrite_timestamp(POOL_CONNECTION_POOL *backend, Node *node,
 			int				 i;
 			PrepareStmt		*p_stmt = (PrepareStmt *) node;
 
-			for (i = 0; i < portal->num_tsparams; i++)
+			for (i = 0; i < pstmt->num_tsparams; i++)
 				p_stmt->argtypes =
 				   	lappend(p_stmt->argtypes, SystemTypeName("timestamptz"));
 		}
